/*
Copyright (c) 2003-2004, Mark Borgerding

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the author nor the names of any contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#undef MIN
#undef MAX

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

/* kiss_fft.h
   defines kiss_fft_scalar as either short or a float type
   and defines
   typedef struct { kiss_fft_scalar r; kiss_fft_scalar i; }kiss_fft_cpx; */
#include "kiss_fft.h"
//#include "math_approx.h"

#define MAXFACTORS 32
/* e.g. an fft of length 128 has 4 factors
 as far as kissfft is concerned
 4*4*4*2
 */

struct kiss_fft_state {
	int nfft;
	int inverse;
	int factors[2 * MAXFACTORS];
	kiss_fft_cpx twiddles[1];
};

/*
  Explanation of macros dealing with complex math:

   C_MUL(m,a,b)         : m = a*b
   C_FIXDIV( c , div )  : if a fixed point impl., c /= div. noop otherwise
   C_SUB( res, a,b)     : res = a - b
   C_SUBFROM( res , a)  : res -= a
   C_ADDTO( res , a)    : res += a
 * */
#ifdef MS_FIXED_POINT
#define FRACBITS 15
#define SAMPPROD ms_int32_t
#define SAMP_MAX 32767

#define SAMP_MIN -SAMP_MAX

#if defined(CHECK_OVERFLOW)
#define CHECK_OVERFLOW_OP(a, op, b)                                                                                    \
	if ((SAMPPROD)(a)op(SAMPPROD)(b) > SAMP_MAX || (SAMPPROD)(a)op(SAMPPROD)(b) < SAMP_MIN) {                          \
		fprintf(stderr, "WARNING:overflow @ " __FILE__ "(%d): (%d " #op " %d) = %ld\n", __LINE__, (a), (b),            \
		        (SAMPPROD)(a)op(SAMPPROD)(b));                                                                         \
	}
#endif

#define smul(a, b) ((SAMPPROD)(a) * (b))
#define sround(x) (kiss_fft_scalar)(((x) + (1 << (FRACBITS - 1))) >> FRACBITS)

#define S_MUL(a, b) sround(smul(a, b))

#define C_MUL(m, a, b)                                                                                                 \
	do {                                                                                                               \
		(m).r = sround(smul((a).r, (b).r) - smul((a).i, (b).i));                                                       \
		(m).i = sround(smul((a).r, (b).i) + smul((a).i, (b).r));                                                       \
	} while (0)

#define C_MUL4(m, a, b)                                                                                                \
	do {                                                                                                               \
		(m).r = PSHR32(smul((a).r, (b).r) - smul((a).i, (b).i), 17);                                                   \
		(m).i = PSHR32(smul((a).r, (b).i) + smul((a).i, (b).r), 17);                                                   \
	} while (0)

#define DIVSCALAR(x, k) (x) = sround(smul(x, SAMP_MAX / k))

#define C_FIXDIV(c, div)                                                                                               \
	do {                                                                                                               \
		DIVSCALAR((c).r, div);                                                                                         \
		DIVSCALAR((c).i, div);                                                                                         \
	} while (0)

#define C_MULBYSCALAR(c, s)                                                                                            \
	do {                                                                                                               \
		(c).r = sround(smul((c).r, s));                                                                                \
		(c).i = sround(smul((c).i, s));                                                                                \
	} while (0)

#else /* not FIXED_POINT*/

#define S_MUL(a, b) ((a) * (b))
#define C_MUL(m, a, b)                                                                                                 \
	do {                                                                                                               \
		(m).r = (a).r * (b).r - (a).i * (b).i;                                                                         \
		(m).i = (a).r * (b).i + (a).i * (b).r;                                                                         \
	} while (0)

#define C_MUL4(m, a, b) C_MUL(m, a, b)

#define C_FIXDIV(c, div) /* NOOP */
#define C_MULBYSCALAR(c, s)                                                                                            \
	do {                                                                                                               \
		(c).r *= (s);                                                                                                  \
		(c).i *= (s);                                                                                                  \
	} while (0)
#endif

#ifndef CHECK_OVERFLOW_OP
#define CHECK_OVERFLOW_OP(a, op, b) /* noop */
#endif

#define C_ADD(res, a, b)                                                                                               \
	do {                                                                                                               \
		CHECK_OVERFLOW_OP((a).r, +, (b).r)                                                                             \
		CHECK_OVERFLOW_OP((a).i, +, (b).i)                                                                             \
		(res).r = (a).r + (b).r;                                                                                       \
		(res).i = (a).i + (b).i;                                                                                       \
	} while (0)
#define C_SUB(res, a, b)                                                                                               \
	do {                                                                                                               \
		CHECK_OVERFLOW_OP((a).r, -, (b).r)                                                                             \
		CHECK_OVERFLOW_OP((a).i, -, (b).i)                                                                             \
		(res).r = (a).r - (b).r;                                                                                       \
		(res).i = (a).i - (b).i;                                                                                       \
	} while (0)
#define C_ADDTO(res, a)                                                                                                \
	do {                                                                                                               \
		CHECK_OVERFLOW_OP((res).r, +, (a).r)                                                                           \
		CHECK_OVERFLOW_OP((res).i, +, (a).i)                                                                           \
		(res).r += (a).r;                                                                                              \
		(res).i += (a).i;                                                                                              \
	} while (0)

#define C_SUBFROM(res, a)                                                                                              \
	do {                                                                                                               \
		CHECK_OVERFLOW_OP((res).r, -, (a).r)                                                                           \
		CHECK_OVERFLOW_OP((res).i, -, (a).i)                                                                           \
		(res).r -= (a).r;                                                                                              \
		(res).i -= (a).i;                                                                                              \
	} while (0)

#ifdef MS_FIXED_POINT
#define KISS_FFT_COS(phase) floor(MIN(32767, MAX(-32767, .5 + 32768 * cos(phase))))
#define KISS_FFT_SIN(phase) floor(MIN(32767, MAX(-32767, .5 + 32768 * sin(phase))))
#define HALF_OF(x) ((x) >> 1)
#elif defined(USE_SIMD)
#define KISS_FFT_COS(phase) _mm_set1_ps(cos(phase))
#define KISS_FFT_SIN(phase) _mm_set1_ps(sin(phase))
#define HALF_OF(x) ((x)*_mm_set1_ps(.5))
#else
#define KISS_FFT_COS(phase) (kiss_fft_scalar) cos(phase)
#define KISS_FFT_SIN(phase) (kiss_fft_scalar) sin(phase)
#define HALF_OF(x) ((x)*.5f)
#endif

#define kf_cexp(x, phase)                                                                                              \
	do {                                                                                                               \
		(x)->r = KISS_FFT_COS(phase);                                                                                  \
		(x)->i = KISS_FFT_SIN(phase);                                                                                  \
	} while (0)

#ifdef MS_FIXED_POINT

#define L1 32767
#define L2 -7651
#define L3 8277
#define L4 -626

static inline ms_word16_t _ms_cos_pi_2(ms_word16_t x) {
	ms_word16_t x2;

	x2 = MULT16_16_P15(x, x);
	return ADD16(
	    1, MIN16(32766, ADD32(SUB16(L1, x2),
	                          MULT16_16_P15(x2, ADD32(L2, MULT16_16_P15(x2, ADD32(L3, MULT16_16_P15(L4, x2))))))));
}

static inline ms_word16_t ms_cos_norm(ms_word32_t x) {
	x = x & 0x0001ffff;
	if (x > SHL32(EXTEND32(1), 16)) x = SUB32(SHL32(EXTEND32(1), 17), x);
	if (x & 0x00007fff) {
		if (x < SHL32(EXTEND32(1), 15)) {
			return _ms_cos_pi_2(EXTRACT16(x));
		} else {
			return NEG32(_ms_cos_pi_2(EXTRACT16(65536 - x)));
		}
	} else {
		if (x & 0x0000ffff) return 0;
		else if (x & 0x0001ffff) return -32767;
		else return 32767;
	}
}

#else
#define ms_cos_norm(x) (cos((.5f * M_PI) * (x)))
#endif

#define kf_cexp2(x, phase)                                                                                             \
	do {                                                                                                               \
		(x)->r = ms_cos_norm((phase));                                                                                 \
		(x)->i = ms_cos_norm((phase)-32768);                                                                           \
	} while (0)

/* a debugging function */
#define pcpx(c) fprintf(stderr, "%g + %gi\n", (double)((c)->r), (double)((c)->i))
